/*
 * Copyright (c) 2009 Sun Microsystems, Inc.  All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * Redistribution of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 
 * Redistribution in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of Sun Microsystems, Inc. or the names of
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * 
 * This software is provided "AS IS," without a warranty of any kind.
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES,
 * INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED.
 * SUN MICROSYSTEMS, INC. ("SUN") AND ITS LICENSORS SHALL NOT BE LIABLE
 * FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.  IN NO EVENT WILL
 * SUN OR ITS LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA,
 * OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR
 * PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE THEORY OF
 * LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE,
 * EVEN IF SUN HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <ctype.h>
#include <sys/time.h>
#include <values.h>

#include <ipmitool/ipmi.h>
#include <ipmitool/ipmi_intf.h>
#include <ipmitool/helper.h>
#include <ipmitool/log.h>
#include <ipmitool/ipmi_sel.h>
#include <ipmitool/ipmi_sdr.h>
#include <ipmitool/ipmi_strings.h>
#include <ipmitool/ipmi_channel.h>
#include <ipmitool/ipmi_sunoem.h>
#include <ipmitool/ipmi_raw.h>

static const struct valstr sunoem_led_type_vals[] = {
	{ 0, "OK2RM" },
	{ 1, "SERVICE" },
	{ 2, "ACT" },
	{ 3, "LOCATE" },
	{ 0xFF, NULL },
};

static const struct valstr sunoem_led_mode_vals[] = {
	{ 0, "OFF" },
	{ 1, "ON" },
	{ 2, "STANDBY" },
	{ 3, "SLOW" },
	{ 4, "FAST" },
	{ 0xFF, NULL },
};
static const struct valstr sunoem_led_mode_optvals[] = {
	{ 0, "STEADY_OFF" },
	{ 1, "STEADY_ON" },
	{ 2, "STANDBY_BLINK" },
	{ 3, "SLOW_BLINK" },
	{ 4, "FAST_BLINK" },
	{ 0xFF, NULL },
};

static const struct valstr sunoem_sensor_mode_vals[] = {
	{ 0, "DEASSERT" },
	{ 1, "ASSERT" },
	{ 0xFF, NULL },
};
#define SUNOEM_SUCCESS 1

/*
 * PRINT_NORMAL: print out the LED value as normal
 * PRINT_ERROR: print out "na" for the LED value
 */
typedef enum {
    PRINT_NORMAL = 0,
    PRINT_ERROR
} print_status_t;
 
int ret_get=0;
int ret_set=0;
static void
ipmi_sunoem_usage(void)
{
	lprintf(LOG_NOTICE, "usage: sunoem <command> [option...]");
	lprintf(LOG_NOTICE, "");
	lprintf(LOG_NOTICE, "   fan speed <auto|0-100>");
	lprintf(LOG_NOTICE, "      Set system fan speed (PWM duty cycle)");
	lprintf(LOG_NOTICE, "");
	lprintf(LOG_NOTICE, "   sshkey set <userid> <id_rsa.pub>");
	lprintf(LOG_NOTICE, "      Set ssh key for a userid into authorized_keys,");
	lprintf(LOG_NOTICE, "      view users with 'user list' command.");
	lprintf(LOG_NOTICE, "");
	lprintf(LOG_NOTICE, "   sshkey del <userid>");
	lprintf(LOG_NOTICE, "      Delete ssh key for userid from authorized_keys,");
	lprintf(LOG_NOTICE, "      view users with 'user list' command.");
	lprintf(LOG_NOTICE, "");
	lprintf(LOG_NOTICE, "   led get <sensorid> [ledtype]");
	lprintf(LOG_NOTICE, "      Read status of LED found in Generic Device Locator.");
	lprintf(LOG_NOTICE, "");
	lprintf(LOG_NOTICE, "   led set <sensorid> <ledmode> [ledtype]");
	lprintf(LOG_NOTICE, "      Set mode of LED found in Generic Device Locator.");
	lprintf(LOG_NOTICE, "");
	lprintf(LOG_NOTICE, "      Use 'sdr list generic' command to get list of Generic");
	lprintf(LOG_NOTICE, "      Devices that are controllable LEDs.");
	lprintf(LOG_NOTICE, "");
	lprintf(LOG_NOTICE, "      Required SIS LED Mode:");
	lprintf(LOG_NOTICE, "         OFF          Off");
	lprintf(LOG_NOTICE, "         ON           Steady On");
	lprintf(LOG_NOTICE, "         STANDBY      100ms on 2900ms off blink rate");
	lprintf(LOG_NOTICE, "         SLOW         1HZ blink rate");
	lprintf(LOG_NOTICE, "         FAST         4HZ blink rate");
	lprintf(LOG_NOTICE, "");
	lprintf(LOG_NOTICE, "      Optional SIS LED Type:");
	lprintf(LOG_NOTICE, "         OK2RM        OK to Remove");
	lprintf(LOG_NOTICE, "         SERVICE      Service Required");
	lprintf(LOG_NOTICE, "         ACT          Activity");
	lprintf(LOG_NOTICE, "         LOCATE       Locate");
	lprintf(LOG_NOTICE, "");
	lprintf(LOG_NOTICE, "   sensor set <sensorid> <mode>");
	lprintf(LOG_NOTICE, "      Set mode of general-purpose output sensor found in");
	lprintf(LOG_NOTICE, "      Compact Sensor record.");
	lprintf(LOG_NOTICE, "");
	lprintf(LOG_NOTICE, "      Use 'sdr list' command to get the current state of");
	lprintf(LOG_NOTICE, "      the general-purpose output sensors.");
	lprintf(LOG_NOTICE, "");
	lprintf(LOG_NOTICE, "      Required sensor mode:");
	lprintf(LOG_NOTICE, "         ASSERT       Assert the sensor");
	lprintf(LOG_NOTICE, "         DEASSERT     Deassert the sensor");
	lprintf(LOG_NOTICE, "");
	lprintf(LOG_NOTICE, "   cli [<command string> ...]");
	lprintf(LOG_NOTICE, "      Execute SP CLI commands.");
	lprintf(LOG_NOTICE, "");
	lprintf(LOG_NOTICE, "   ping NUMBER <q>");
	lprintf(LOG_NOTICE, "      Send and Receive NUMBER (64 Byte) packets.");
	lprintf(LOG_NOTICE, "      q - Quiet. Displays output at start and end");
	lprintf(LOG_NOTICE, "");
	lprintf(LOG_NOTICE, "   version");
	lprintf(LOG_NOTICE, "      Display the software version");
	lprintf(LOG_NOTICE, "");
	lprintf(LOG_NOTICE, "   nacname <ipmi nac name>");
	lprintf(LOG_NOTICE, "      Returns the full nac name");
	lprintf(LOG_NOTICE, "");
	lprintf(LOG_NOTICE, "   getval <target name>");
	lprintf(LOG_NOTICE, "      Returns the ILOM property value");
	lprintf(LOG_NOTICE, "");
}

#define SUNOEM_FAN_MODE_AUTO    0x00
#define SUNOEM_FAN_MODE_MANUAL  0x01

/* 
 * IPMI Request Data: 1 byte
 *
 * [byte 0]  FanMode    Fan mode, 0x00 = auto, 0x01 = manual
 */
static int
ipmi_sunoem_fan_mode(struct ipmi_intf * intf, uint8_t modeb)
{
	struct ipmi_rs * rsp;
	struct ipmi_rq req;

	bzero (&req, sizeof(req));

	req.msg.netfn = IPMI_NETFN_SUNOEM;
	req.msg.cmd = IPMI_SUNOEM_SET_FAN_MODE;
        
        switch (modeb) {
        case SUNOEM_FAN_MODE_AUTO:
        case SUNOEM_FAN_MODE_MANUAL:
            req.msg.data = &modeb;
            break;
        default:
            lprintf(LOG_NOTICE, "Invalid fan mode");
            return -1;
        }

	req.msg.data_len = 1;
		
	rsp = intf->sendrecv(intf, &req);
	if (rsp == NULL) {
		lprintf(LOG_ERR, "Sun OEM Set Fan Mode command failed");
		return -1;
	}
	else if (rsp->ccode > 0) {
		lprintf(LOG_ERR, "Sun OEM Set Fan Mode command failed: %s",
			val2str(rsp->ccode, completion_code_vals));
		return -1;
	}

	return 0;
}

/* 
 * IPMI Request Data: 1 byte
 *
 * [byte 0]  FanSpeed    Fan speed as percentage
 */
static int
ipmi_sunoem_fan_speed(struct ipmi_intf * intf, char *arg)
{
	struct ipmi_rs * rsp;
	struct ipmi_rq req;
        uint8_t speed;
        int rc = 0;

	bzero (&req, sizeof(req));

	/* 
	 * sunoem fan speed <arg>
	 */

        if (strcmp(arg, "auto") == 0) {
            if ((rc = ipmi_sunoem_fan_mode(intf, SUNOEM_FAN_MODE_AUTO)) < 0) {
                goto finish;
            }

            goto finish;
        } else {
            speed = strtol(arg, NULL, 0);
        }

        if ((rc = ipmi_sunoem_fan_mode(intf, SUNOEM_FAN_MODE_MANUAL)) < 0) {
            goto finish;
        }

	if (speed > 100) {
		lprintf(LOG_NOTICE, "Invalid fan speed: %d", speed);
		return -1;
	}

	req.msg.netfn = IPMI_NETFN_SUNOEM;
	req.msg.cmd = IPMI_SUNOEM_SET_FAN_SPEED;
	req.msg.data = &speed;
	req.msg.data_len = 1;
		
	rsp = intf->sendrecv(intf, &req);
	if (rsp == NULL) {
		lprintf(LOG_ERR, "Sun OEM Set Fan Speed command failed");
		return -1;
	}
	else if (rsp->ccode > 0) {
		lprintf(LOG_ERR, "Sun OEM Set Fan Speed command failed: %s",
			val2str(rsp->ccode, completion_code_vals));
		return -1;
	}

	printf("Set Fan speed to %d%%\n", speed);

finish:
	return rc;
}

static void
__sdr_list_empty(struct sdr_record_list * head)
{
	struct sdr_record_list * e, * f;
	for (e = head; e != NULL; e = f) {
		f = e->next;
		free(e);
	}
	head = NULL;
}

/*
 * led_print
 *  Print out the led name and the state, if stat is PRINT_NORMAL.
 *  Otherwise, print out the led name and "na".
 *  The state parameter is not referenced if stat is not PRINT_NORMAL.
 */
static void
led_print(const char * name, print_status_t stat, uint8_t state)
{
    const char *theValue;

    if (stat == PRINT_NORMAL) {
        theValue = val2str(state, sunoem_led_mode_vals);
    } else {
        theValue = "na";
    }

    if (csv_output) { 
        printf("%s,%s\n", name, theValue);
    } else { 
        printf("%-16s | %s\n", name, theValue);
    }
}

struct ipmi_rs *
sunoem_led_get(struct ipmi_intf * intf,
		 struct sdr_record_generic_locator * dev,
		 int ledtype)
{
	struct ipmi_rs * rsp;
	struct ipmi_rq req;
	uint8_t rqdata[7];
	int rqdata_len;

	if (dev == NULL)
		return NULL;

	rqdata[0] = dev->dev_slave_addr;
	if (ledtype == 0xFF)
		rqdata[1] = dev->oem;
	else
		rqdata[1] = ledtype;
	rqdata[2] = dev->dev_access_addr;
	rqdata[3] = dev->oem;
		rqdata[4] = dev->entity.id;
		rqdata[5] = dev->entity.instance;
		rqdata[6] = 0;
		rqdata_len = 7;

	req.msg.netfn = IPMI_NETFN_SUNOEM;
	req.msg.cmd = IPMI_SUNOEM_LED_GET;
	req.msg.lun = dev->lun;
	req.msg.data = rqdata;
	req.msg.data_len = rqdata_len;
		
	rsp = intf->sendrecv(intf, &req);
    /*
     * Just return NULL if there was
     * an error.
     */
    if (rsp == NULL) {
        return NULL;
    } else if (rsp->ccode > 0) {
        return NULL;
    } else {
        return rsp;
    }
}

struct ipmi_rs *
sunoem_led_set(struct ipmi_intf * intf,
		 struct sdr_record_generic_locator * dev,
		 int ledtype, int ledmode)
{
	struct ipmi_rs * rsp;
	struct ipmi_rq req;
	uint8_t rqdata[9];
	int rqdata_len;

	if (dev == NULL)
		return NULL;

	rqdata[0] = dev->dev_slave_addr;
	if (ledtype == 0xFF)
		rqdata[1] = dev->oem;
	else
		rqdata[1] = ledtype;
	rqdata[2] = dev->dev_access_addr;
	rqdata[3] = dev->oem;
	rqdata[4] = ledmode;
		rqdata[5] = dev->entity.id;
		rqdata[6] = dev->entity.instance;
		rqdata[7] = 0;
		rqdata[8] = 0;
		rqdata_len = 9;

	req.msg.netfn = IPMI_NETFN_SUNOEM;
	req.msg.cmd = IPMI_SUNOEM_LED_SET;
	req.msg.lun = dev->lun;
	req.msg.data = rqdata;
	req.msg.data_len = rqdata_len;
		
    rsp = intf->sendrecv(intf, &req);
    if (rsp == NULL) {
        lprintf(LOG_ERR, "Sun OEM Set LED command failed");
        return NULL;
    }
    else if (rsp->ccode > 0) {
        lprintf(LOG_ERR, "Sun OEM Set LED command failed: %s",
            val2str(rsp->ccode, completion_code_vals));
        return NULL;
    }

    return rsp;
}

static void
sunoem_led_get_byentity(struct ipmi_intf * intf, uint8_t entity_id,
			  uint8_t entity_inst, int ledtype)
{
	struct ipmi_rs * rsp;
	struct sdr_record_list *elist, *e;
	struct entity_id entity;

	if (entity_id == 0)
		return;

	/* lookup sdrs with this entity */
	memset(&entity, 0, sizeof(struct entity_id));
	entity.id = entity_id;
	entity.instance = entity_inst;

	elist = ipmi_sdr_find_sdr_byentity(intf, &entity);
    
	if (elist == NULL)
		ret_get=-1;			//Fixed for bug No:6812256

	/* for each generic sensor get its led state */
	for (e = elist; e != NULL; e = e->next) {
		if (e->type != SDR_RECORD_TYPE_GENERIC_DEVICE_LOCATOR)
			continue;
		rsp = sunoem_led_get(intf, e->record.genloc, ledtype);
		if (rsp && rsp->data_len == 1) {
			led_print((const char *)e->record.genloc->id_string, PRINT_NORMAL, rsp->data[0]);
		} else {
			led_print((const char *)e->record.genloc->id_string, PRINT_ERROR, 0);
			if (rsp == NULL)
				ret_get=-1;			//Fixed for bug No:6812256
        }
    }

	__sdr_list_empty(elist);
}

static void
sunoem_led_set_byentity(struct ipmi_intf * intf, uint8_t entity_id,
			  uint8_t entity_inst, int ledtype, int ledmode)
{
	struct ipmi_rs * rsp;
	struct sdr_record_list *elist, *e;
	struct entity_id entity;

	if (entity_id == 0)
		return;

	/* lookup sdrs with this entity */
	memset(&entity, 0, sizeof(struct entity_id));
	entity.id = entity_id;
	entity.instance = entity_inst;

	elist = ipmi_sdr_find_sdr_byentity(intf, &entity);
    
	if (elist == NULL)			//Fixed for bug No:6812256
		ret_set=-1;

	/* for each generic sensor set its led state */
	for (e = elist; e != NULL; e = e->next) {
		if (e->type != SDR_RECORD_TYPE_GENERIC_DEVICE_LOCATOR)
			continue;
		rsp = sunoem_led_set(intf, e->record.genloc, ledtype, ledmode);
		if (rsp && rsp->data_len == 0) {
			led_print((const char *)e->record.genloc->id_string, PRINT_NORMAL, ledmode);
		}
		else if(rsp == NULL){			//Fixed for bug No:6812256
			ret_set=-1;
		}
	}

	__sdr_list_empty(elist);
}

/* 
 * IPMI Request Data: 5 bytes
 *
 * [byte 0]  devAddr     Value from the "Device Slave Address" field in
 *                       LED's Generic Device Locator record in the SDR
 * [byte 1]  led         LED Type: OK2RM, ACT, LOCATE, SERVICE
 * [byte 2]  ctrlrAddr   Controller address; value from the "Device
 *                       Access Address" field, 0x20 if the LED is local
 * [byte 3]  hwInfo      The OEM field from the SDR record
 * [byte 4]  force       1 = directly access the device
 *                       0 = go thru its controller
 *                       Ignored if LED is local
 * 
 * The format below is for Sun Blade Modular systems only
 * [byte 4]  entityID    The entityID field from the SDR record
 * [byte 5]  entityIns   The entityIns field from the SDR record
 * [byte 6]  force       1 = directly access the device
 *                       0 = go thru its controller
 *                       Ignored if LED is local
 */
static int
ipmi_sunoem_led_get(struct ipmi_intf * intf,  int argc, char ** argv)
{
	struct ipmi_rs * rsp;
	struct sdr_record_list *sdr;
	struct sdr_record_list *alist, *a;
	struct sdr_record_entity_assoc *assoc;
	int ledtype = 0xFF;
	int i;

	/* 
	 * sunoem led/sbled get <id> [type]
	 */

	if (argc < 1 || strncmp(argv[0], "help", 4) == 0) {
		ipmi_sunoem_usage();
		return 0;
	}

	if (argc > 1) {
		ledtype = str2val(argv[1], sunoem_led_type_vals);
		if (ledtype == 0xFF) 
			lprintf(LOG_ERR, "Unknow ledtype, will use data from the SDR oem field");
	}

	if (strncasecmp(argv[0], "all", 3) == 0) {
		/* do all generic sensors */
		alist = ipmi_sdr_find_sdr_bytype(intf, SDR_RECORD_TYPE_GENERIC_DEVICE_LOCATOR);
		
		if (alist == NULL)			//Fixed for bug No:6812256
			return -1;

		for (a = alist; a != NULL; a = a->next) {
			if (a->type != SDR_RECORD_TYPE_GENERIC_DEVICE_LOCATOR)
				continue;
			if (a->record.genloc->entity.logical)
				continue;
			rsp = sunoem_led_get(intf, a->record.genloc, ledtype);
			if (rsp && rsp->data_len == 1) {
				led_print((const char *)a->record.genloc->id_string, PRINT_NORMAL, rsp->data[0]);
			} else {
				led_print((const char *)a->record.genloc->id_string, PRINT_ERROR, 0);
				if (rsp == NULL)			//Fixed for bug No:6812256
				    ret_get=-1;
            }
		}
		__sdr_list_empty(alist);
		
		if (ret_get == -1)			//Fixed for bug No:6812256
			return -1;

		return 0;
	}

	/* look up generic device locator record in SDR */
	sdr = ipmi_sdr_find_sdr_byid(intf, argv[0]);

	if (sdr == NULL) {
		lprintf(LOG_ERR, "No Sensor Data Record found for %s", argv[0]);
		return -1;
	}

	if (sdr->type != SDR_RECORD_TYPE_GENERIC_DEVICE_LOCATOR) {
		lprintf(LOG_ERR, "Invalid SDR type %d", sdr->type);
		return -1;
	}

	if (!sdr->record.genloc->entity.logical) {
		/*
		 * handle physical entity
		 */
		rsp = sunoem_led_get(intf, sdr->record.genloc, ledtype);
		if (rsp && rsp->data_len == 1) {
			led_print((const char *)sdr->record.genloc->id_string, PRINT_NORMAL, rsp->data[0]);
		} else {
			led_print((const char *)sdr->record.genloc->id_string, PRINT_ERROR, 0);
			if (rsp == NULL)
				ret_get=-1;			//Fixed for bug No:6812256
        }
		
		if (ret_get == -1)
			return -1;			//Fixed for bug No:6812256
		
		return 0;
	}

	/* 
	 * handle logical entity for LED grouping
	 */

	lprintf(LOG_INFO, "LED %s is logical device", argv[0]);

	/* get entity assoc records */
	alist = ipmi_sdr_find_sdr_bytype(intf, SDR_RECORD_TYPE_ENTITY_ASSOC);
	
	if (alist == NULL)			//Fixed for bug No:6812256
			return -1;

	for (a = alist; a != NULL; a = a->next) {
		if (a->type != SDR_RECORD_TYPE_ENTITY_ASSOC)
			continue;
		assoc = a->record.entassoc;
		if (assoc == NULL)
			continue;

		/* check that the entity id/instance matches our generic record */
		if (assoc->entity.id != sdr->record.genloc->entity.id ||
		    assoc->entity.instance != sdr->record.genloc->entity.instance)
			continue;

		if (assoc->flags.isrange) {
			/*
			 * handle ranged entity associations
			 *
			 * the test for non-zero entity id is handled in
			 * sunoem_led_get_byentity()
			 */

			/* first range set - id 1 and 2 must be equal */
			if (assoc->entity_id_1 == assoc->entity_id_2)
				for (i = assoc->entity_inst_1; i <= assoc->entity_inst_2; i++)
					sunoem_led_get_byentity(intf, assoc->entity_id_1, i, ledtype);

			/* second range set - id 3 and 4 must be equal */
			if (assoc->entity_id_3 == assoc->entity_id_4)
				for (i = assoc->entity_inst_3; i <= assoc->entity_inst_4; i++)
					sunoem_led_get_byentity(intf, assoc->entity_id_3, i, ledtype);
		}
		else {
			/*
			 * handle entity list
			 */
			sunoem_led_get_byentity(intf, assoc->entity_id_1,
						assoc->entity_inst_1, ledtype);
			sunoem_led_get_byentity(intf, assoc->entity_id_2,
						assoc->entity_inst_2, ledtype);
			sunoem_led_get_byentity(intf, assoc->entity_id_3,
						assoc->entity_inst_3, ledtype);
			sunoem_led_get_byentity(intf, assoc->entity_id_4,
						assoc->entity_inst_4, ledtype);
		}
	}

	__sdr_list_empty(alist);

	if (ret_get == -1)			//Fixed for bug No:6812256
		return -1;

	return 0;
}

/* 
 * IPMI Request Data: 7 bytes
 *
 * [byte 0]  devAddr     Value from the "Device Slave Address" field in
 *                       LED's Generic Device Locator record in the SDR
 * [byte 1]  led         LED Type: OK2RM, ACT, LOCATE, SERVICE
 * [byte 2]  ctrlrAddr   Controller address; value from the "Device
 *                       Access Address" field, 0x20 if the LED is local
 * [byte 3]  hwInfo      The OEM field from the SDR record
 * [byte 4]  mode        LED Mode: OFF, ON, STANDBY, SLOW, FAST
 * [byte 5]  force       TRUE - directly access the device
 *                       FALSE - go thru its controller
 *                       Ignored if LED is local
 * [byte 6]  role        Used by BMC for authorization purposes
 *
 * The format below is for Sun Blade Modular systems only
 * [byte 5]  entityID    The entityID field from the SDR record
 * [byte 6]  entityIns   The entityIns field from the SDR record
 * [byte 7]  force       TRUE - directly access the device
 *                       FALSE - go thru its controller
 *                       Ignored if LED is local
 * [byte 8]  role        Used by BMC for authorization purposes
 *
 *
 * IPMI Response Data: 1 byte
 * 
 * [byte 0]  mode     LED Mode: OFF, ON, STANDBY, SLOW, FAST
 */

static int
ipmi_sunoem_led_set(struct ipmi_intf * intf,  int argc, char ** argv)
{
	struct ipmi_rs * rsp;
	struct sdr_record_list *sdr;
	struct sdr_record_list *alist, *a;
	struct sdr_record_entity_assoc *assoc;
	int ledmode;
	int ledtype = 0xFF;
	int i;

	/* 
	 * sunoem led/sbled set <id> <mode> [type]
	 */

	if (argc < 2 || strncmp(argv[0], "help", 4) == 0) {
		ipmi_sunoem_usage();
		return 0;
	}

	ledmode = str2val(argv[1], sunoem_led_mode_vals);
	if (ledmode == 0xFF) {
		ledmode = str2val(argv[1], sunoem_led_mode_optvals);
		if (ledmode == 0xFF) {
			lprintf(LOG_NOTICE, "Invalid LED Mode: %s", argv[1]);
			return -1;
		}
	}

	if (argc > 3) {
		ledtype = str2val(argv[2], sunoem_led_type_vals);
		if (ledtype == 0xFF) 
			lprintf(LOG_ERR, "Unknow ledtype, will use data from the SDR oem field");
	}

	if (strncasecmp(argv[0], "all", 3) == 0) {
		/* do all generic sensors */
		alist = ipmi_sdr_find_sdr_bytype(intf, SDR_RECORD_TYPE_GENERIC_DEVICE_LOCATOR);
	
		if (alist == NULL)			//Fixed for bug No:6812256
			return -1;

		for (a = alist; a != NULL; a = a->next) {
			if (a->type != SDR_RECORD_TYPE_GENERIC_DEVICE_LOCATOR)
				continue;
			if (a->record.genloc->entity.logical)
				continue;
			rsp = sunoem_led_set(intf, a->record.genloc, ledtype, ledmode);
			if (rsp && rsp->ccode == 0) {
				led_print((const char *)a->record.genloc->id_string, PRINT_NORMAL, ledmode);
			}
			else if (rsp == NULL){			//Fixed for bug No:6812256
				ret_set=-1;
			}
		}
		__sdr_list_empty(alist);
		
		if (ret_set == -1)			//Fixed for bug No:6812256
			return -1;

		return 0;
	}

	/* look up generic device locator records in SDR */
	sdr = ipmi_sdr_find_sdr_byid(intf, argv[0]);

	if (sdr == NULL) {
		lprintf(LOG_ERR, "No Sensor Data Record found for %s",
			argv[0]);
		return -1;
	}

	if (sdr->type != SDR_RECORD_TYPE_GENERIC_DEVICE_LOCATOR) {
		lprintf(LOG_ERR, "Invalid SDR type %d", sdr->type);
		return -1;
	}

	if (!sdr->record.genloc->entity.logical) {
		/*
		 * handle physical entity
		 */
		rsp = sunoem_led_set(intf, sdr->record.genloc, ledtype, ledmode);
		if (rsp && rsp->ccode == 0) {
			led_print(argv[0], PRINT_NORMAL, ledmode);
		}
		if (rsp == NULL)			//Fixed for bug No:6812256
			return -1;

		return 0;
	}

	/* 
	 * handle logical entity for LED grouping
	 */

	lprintf(LOG_INFO, "LED %s is logical device", argv[0]);

	/* get entity assoc records */
	alist = ipmi_sdr_find_sdr_bytype(intf, SDR_RECORD_TYPE_ENTITY_ASSOC);
	
	if (alist == NULL)			//Fixed for bug No:6812256
			return -1;

	for (a = alist; a != NULL; a = a->next) {
		if (a->type != SDR_RECORD_TYPE_ENTITY_ASSOC)
			continue;
		assoc = a->record.entassoc;
		if (assoc == NULL)
			continue;

		/* check that the entity id/instance matches our generic record */
		if (assoc->entity.id != sdr->record.genloc->entity.id ||
		    assoc->entity.instance != sdr->record.genloc->entity.instance)
			continue;

		if (assoc->flags.isrange) {
			/*
			 * handle ranged entity associations
			 *
			 * the test for non-zero entity id is handled in
			 * sunoem_led_get_byentity()
			 */

			/* first range set - id 1 and 2 must be equal */
			if (assoc->entity_id_1 == assoc->entity_id_2)
				for (i = assoc->entity_inst_1; i <= assoc->entity_inst_2; i++)
					sunoem_led_set_byentity(intf, assoc->entity_id_1, i, ledtype, ledmode);

			/* second range set - id 3 and 4 must be equal */
			if (assoc->entity_id_3 == assoc->entity_id_4)
				for (i = assoc->entity_inst_3; i <= assoc->entity_inst_4; i++)
					sunoem_led_set_byentity(intf, assoc->entity_id_3, i, ledtype, ledmode);
		}
		else {
			/*
			 * handle entity list
			 */
			sunoem_led_set_byentity(intf, assoc->entity_id_1,
						assoc->entity_inst_1, ledtype, ledmode);
			sunoem_led_set_byentity(intf, assoc->entity_id_2,
						assoc->entity_inst_2, ledtype, ledmode);
			sunoem_led_set_byentity(intf, assoc->entity_id_3,
						assoc->entity_inst_3, ledtype, ledmode);
			sunoem_led_set_byentity(intf, assoc->entity_id_4,
						assoc->entity_inst_4, ledtype, ledmode);
		}
	}

	__sdr_list_empty(alist);
	
	if (ret_set == -1)			//Fixed for bug No:6812256
			return -1;

	return 0;
}

static int
ipmi_sunoem_sshkey_del(struct ipmi_intf * intf, uint8_t uid)
{
	struct ipmi_rs * rsp;
	struct ipmi_rq req;

	memset(&req, 0, sizeof(struct ipmi_rq));
	req.msg.netfn = IPMI_NETFN_SUNOEM;
	req.msg.cmd = IPMI_SUNOEM_DEL_SSH_KEY;
	req.msg.data = &uid;
	req.msg.data_len = 1;

	rsp = intf->sendrecv(intf, &req);
	if (rsp == NULL) {
		lprintf(LOG_ERR, "Unable to delete ssh key for UID %d", uid);
		return -1;
	}
	else if (rsp->ccode > 0) {
		lprintf(LOG_ERR, "Unable to delete ssh key for UID %d: %s", uid,
			val2str(rsp->ccode, completion_code_vals));
		return -1;
	}

	printf("Deleted SSH key for user id %d\n", uid);
	return 0;
}

#define SSHKEY_BLOCK_SIZE	64
static int
ipmi_sunoem_sshkey_set(struct ipmi_intf * intf, uint8_t uid, char * ifile)
{
	struct ipmi_rs * rsp;
	struct ipmi_rq req;
	FILE * fp;
	int count;
	uint16_t i_size, r, size;
	uint8_t wbuf[SSHKEY_BLOCK_SIZE + 3];

	if (ifile == NULL) {
		lprintf(LOG_ERR, "Invalid or misisng input filename");
		return -1;
	}

	fp = ipmi_open_file_read(ifile);
	if (fp == NULL) {
		lprintf(LOG_ERR, "Unable to open file %s for reading", ifile);
		return -1;
	}

	printf("Setting SSH key for user id %d...", uid);

	memset(&req, 0, sizeof(struct ipmi_rq));
	req.msg.netfn = IPMI_NETFN_SUNOEM;
	req.msg.cmd = IPMI_SUNOEM_SET_SSH_KEY;
	req.msg.data = wbuf;

	fseek(fp, 0, SEEK_END);
	size = ftell(fp);
	fseek(fp, 0, SEEK_SET);

	for (r = 0; r < size; r += i_size) {
		i_size = size - r;
		if (i_size > SSHKEY_BLOCK_SIZE)
			i_size = SSHKEY_BLOCK_SIZE;

		memset(wbuf, 0, SSHKEY_BLOCK_SIZE);
		fseek(fp, r, SEEK_SET);
		count = fread(wbuf+3, 1, i_size, fp);
		if (count != i_size) {
			lprintf(LOG_ERR, "Unable to read %d bytes from file %s", i_size, ifile);
			fclose(fp);
			return -1;
		}

		printf(".");
		fflush(stdout);

		wbuf[0] = uid;
		if ((r + SSHKEY_BLOCK_SIZE) >= size)
			wbuf[1] = 0xff;
		else
			wbuf[1] = (uint8_t)(r / SSHKEY_BLOCK_SIZE);
		wbuf[2] = i_size;
		req.msg.data_len = i_size + 3;

		rsp = intf->sendrecv(intf, &req);
		if (rsp == NULL) {
			lprintf(LOG_ERR, "Unable to set ssh key for UID %d", uid);
			break;
		}
	}

	printf("done\n");

	fclose(fp);
	return 0;
}


static void
oemsensor_print(const char * name, uint8_t state)
{
	if (csv_output)
		printf("%s,%s\n", name, val2str(state, sunoem_sensor_mode_vals));
	else
		printf("%-16s | %s\n", name, val2str(state, sunoem_sensor_mode_vals));
}

struct ipmi_rs *
sunoem_sensor_set(struct ipmi_intf * intf,
		 struct sdr_record_compact_sensor * dev,
		 int sensormode)
{
	struct ipmi_rs * rsp;
	struct ipmi_rq req;
	uint8_t rqdata[7];
	int rqdata_len = 7;

	if (dev == NULL)
		return NULL;

	rqdata[0] = dev->keys.owner_id;
	rqdata[1] = dev->keys.sensor_num;
	rqdata[2] = sensormode;
	rqdata[3] = dev->entity.id;
	rqdata[4] = dev->entity.instance;
	rqdata[5] = 0;
	rqdata[6] = 0;

	req.msg.netfn = IPMI_NETFN_SUNOEM;
	req.msg.cmd = IPMI_SUNOEM_SENSOR_SET;
	req.msg.lun = dev->keys.lun;
	req.msg.data = rqdata;
	req.msg.data_len = rqdata_len;
		
	rsp = intf->sendrecv(intf, &req);
	if (rsp == NULL) {
		lprintf(LOG_ERR, "Sun OEM Set Sensor command failed");
		return NULL;
	}
	else if (rsp->ccode > 0) {
		lprintf(LOG_ERR, "Sun OEM Set Sensor command failed: %s (0x%02x)",
			val2str(rsp->ccode, completion_code_vals), rsp->ccode);
		return NULL;
	}

	return rsp;
}

/* 
 * IPMI Request Data: 7 bytes
 *
 * [byte 0]  ownerID     Value from the "Sensor Owner ID" field in
 *                       sensor's Compact Sensor Record in the SDR
 * [byte 1]  sensorNum   The Sensor Number field from the SDR record
 * [byte 2]  mode        ASSERT or DEASSERT
 * [byte 3]  entityID    The entityID field from the SDR record
 * [byte 4]  entityIns   The entityIns field from the SDR record
 * [byte 5]  force       TRUE - directly access the device
 *                       FALSE - go thru its controller
 *                       Ignored if sensor is local
 * [byte 6]  role        Used by BMC for authorization purposes
 *
 * IPMI Response Data: 1 byte
 * 
 * [byte 0]  mode        ASSERT, DEASSERT
 */

static int
ipmi_sunoem_sensor_set(struct ipmi_intf * intf,  int argc, char ** argv)
{
	struct ipmi_rs * rsp;
	struct sdr_record_list *sdr;
	int sensormode;

	/* 
	 * sunoem sensor set <id> <mode> [type]
	 */

	if (argc < 2 || strncmp(argv[0], "help", 4) == 0) {
		ipmi_sunoem_usage();
		return 0;
	}

	sensormode = str2val(argv[1], sunoem_sensor_mode_vals);
	if (sensormode == 0xFF) {
		lprintf(LOG_NOTICE, "Invalid sensor mode: %s", argv[1]);
		return -1;
	}

	/* look up sensor record in SDR */
	sdr = ipmi_sdr_find_sdr_byid(intf, argv[0]);

	if (sdr == NULL) {
		lprintf(LOG_ERR, "No Sensor Data Record found for %s",
			argv[0]);
		return -1;
	}

	if (sdr->type != SDR_RECORD_TYPE_COMPACT_SENSOR) {
		lprintf(LOG_ERR, "Invalid SDR type %d", sdr->type);
		return -1;
	}

	rsp = sunoem_sensor_set(intf, sdr->record.compact, sensormode);
	
	if (rsp == NULL)			//Fixed for bug No:6812256
		return -1;

	if (rsp && rsp->ccode == 0) {
		oemsensor_print(argv[0], sensormode);
	}

	return 0;
}

/*
 * This structure is used in both the request to and response from the BMC.
 */
#define SUNOEM_CLI_VERSION       1
#define SUNOEM_CLI_HEADER        8 /* command + spare + handle */
#define SUNOEM_CLI_BUF_SIZE      (80 - SUNOEM_CLI_HEADER) /* Total 80 bytes */
#define SUNOEM_CLI_MSG_SIZE(msg) (SUNOEM_CLI_HEADER + strlen((msg).buf) + 1)
typedef struct {
    /*
     * Set version to SUNOEM_CLI_VERSION.
     */
    uint8_t version;
    /*
     * The command in a request, or in a response indicates an error if
     * non-zero.
     */
    uint8_t command_response;
    uint8_t spare[2];
    /*
     * Opaque 4-byte handle, supplied in the response to an OPEN request,
     * and used in all subsequent POLL and CLOSE requests.
     */
    uint8_t handle[4];
    /*
     * The client data in a request, or the server data in a response. Must
     * by null terminated, i.e., it must be at least one byte, but can be
     * smaller if there's less data.
     */
    char buf[SUNOEM_CLI_BUF_SIZE];
} __attribute__((packed)) sunoem_cli_msg_t;

/*
 * Command codes for the command_request field in each request.
 */
#define SUNOEM_CLI_CMD_OPEN   0 /* Open a new connection */
#define SUNOEM_CLI_CMD_FORCE  1 /* Close any existing connection, then open */
#define SUNOEM_CLI_CMD_CLOSE  2 /* Close the current connection */
#define SUNOEM_CLI_CMD_POLL   3 /* Poll for new data to/from the server */
#define SUNOEM_CLI_CMD_EOF    4 /* Poll, client is out of data */

#define SUNOEM_CLI_MAX_RETRY  3 /* Maximum number of retries */
static int
ipmi_sunoem_cli(struct ipmi_intf * intf, int argc, char *argv[])
{
    struct ipmi_rs *rsp;
    struct ipmi_rq req;
    sunoem_cli_msg_t cli_req;
    sunoem_cli_msg_t *cli_rsp;
    int arg_num = 0;
    int arg_pos = 0;
    time_t wait_time = 0;
    int retries;

    /* Prepare to open an SP shell session */
    memset(&cli_req, 0, sizeof (cli_req));
    cli_req.version = SUNOEM_CLI_VERSION;
    cli_req.command_response = SUNOEM_CLI_CMD_OPEN;
    if (argc > 0 && strcmp(argv[0], "force") == 0) {
        cli_req.command_response = SUNOEM_CLI_CMD_FORCE;
        argc--;
        argv++;
    }
    memset(&req, 0, sizeof (req));
    req.msg.netfn = IPMI_NETFN_SUNOEM;
    req.msg.cmd = IPMI_SUNOEM_CLI;
    req.msg.data = (uint8_t *)&cli_req;
    req.msg.data_len = SUNOEM_CLI_HEADER + 1;
    rsp = intf->sendrecv(intf, &req);
    if (rsp == NULL || rsp->ccode != 0) {
        lprintf(LOG_ERR, "Sun OEM cli command failed");
        return (-1);
    }
    cli_rsp = (sunoem_cli_msg_t *)rsp->data;
    if (cli_rsp->command_response != 0) {
        lprintf(LOG_ERR, "Failed to connect: %s", cli_rsp->buf);
        return (-1);
    }
    printf("Connected. Use ^D to exit.\n");
    fflush(NULL);

    /*
     * Remember the handle provided in the response, and issue a
     * series of "poll" commands to send and get data
     */
    memcpy(cli_req.handle, cli_rsp->handle, 4);
    cli_req.command_response = SUNOEM_CLI_CMD_POLL;
    while (rsp->ccode == 0 && cli_rsp->command_response == 0) {
        int count = 0;
        cli_req.buf[0] = '\0';
        if (argc == 0) {
            /*
             * Accept input from stdin. Use select so we don't hang if
             * there's no input to read. Select timeout is 500 msec.
             */
            struct timeval tv = { 0, 500000 }; /* 500 msec */
            fd_set rfds;
            FD_ZERO(&rfds);
            FD_SET(0, &rfds);
            int rc = select(1, &rfds, NULL, NULL, &tv);
            if (rc < 0) {
                /* Select returned an error so close and exit */
                printf("Broken pipe\n");
                cli_req.command_response = SUNOEM_CLI_CMD_CLOSE;
            } else if (rc > 0) {
                /* Read data from stdin */
                count = read(0, cli_req.buf, sizeof (cli_req.buf) - 1);
                if (count <= 0) {
                    /*
                     * Select said there was data but there was nothing to
                     * read. This implies user hit ^D.
                     */
                    cli_req.command_response = SUNOEM_CLI_CMD_EOF;
                    count = 0;
                }
            }
        } else {
            /* Get data from command line arguments */
            time_t now = time(NULL);
            if (now < wait_time) {
                /* Do nothing; we're waiting */
            } else if (arg_num >= argc) {
                /* Last arg was sent. Set EOF */
                cli_req.command_response = SUNOEM_CLI_CMD_EOF;
            } else if (strncmp(argv[arg_num], "@wait", 5) == 0) {
                /* This is a wait command */
                char *s = &argv[arg_num][5];
                if (*s == '=') {
                    /* Allow the user to specify @wait=5 */
                    s++;
                }
                time_t delay = strtol(s, NULL, 0);
                wait_time = now + delay;
                arg_num++;
            } else {
                /*
                 * Take data from args. It may be that the argument is larger
                 * than the request buffer can hold. So pull off BUF_SIZE
                 * number of characters at a time. When we've consumed the
                 * entire arg, append a newline and advance to the next arg.
                 */
                int i;
                char *s = argv[arg_num];
                for (i = arg_pos;
                     s[i] != '\0' && count < (SUNOEM_CLI_BUF_SIZE - 2);
                     i++, count++) {
                    cli_req.buf[count] = s[i];
                }
                if (s[i] == '\0') {
                    /* Reached end of the arg string, so append a newline */
                    cli_req.buf[count++] = '\n';
                    /* Reset pos to 0 and advance to the next arg next time */
                    arg_pos = 0;
                    arg_num++;
                } else {
                    /*
                     * Otherwise, there's still more characters in the arg
                     * to send, so remember where we left off
                     */
                    arg_pos = i;
                }
            }
        }
        /*
         * Now send the clients's data (if any) and get data back from the
         * server. Loop while the server is giving us data until we suck
         * it dry.
         */
        do {
            cli_req.buf[count++] = '\0'; /* Terminate the string */
            memset(&req, 0, sizeof (req));
            req.msg.netfn = IPMI_NETFN_SUNOEM;
            req.msg.cmd = 0x19;
            req.msg.data = (uint8_t *)&cli_req;
            req.msg.data_len = SUNOEM_CLI_HEADER + count;
            for (retries = 0; retries <= SUNOEM_CLI_MAX_RETRY; retries++) {
                rsp = intf->sendrecv(intf, &req);
                if (rsp == NULL) {
                    lprintf(LOG_ERR, "Communication error");
                    return (-1);
                }
                if (rsp->ccode == IPMI_CC_TIMEOUT) { /* Retry if timed out. */
                    if (retries == SUNOEM_CLI_MAX_RETRY) { /* If it's the last retry. */
                        lprintf(LOG_ERR, "Excessive timeout");
                        return (-1);
                    }
                    continue;
                }
                break;
			}

            cli_rsp = (sunoem_cli_msg_t *)rsp->data;
            /* Make sure response string is null terminated */
            cli_rsp->buf[sizeof(cli_rsp->buf) - 1] = '\0';
            printf("%s", cli_rsp->buf);
            fflush(NULL); /* Flush partial lines to stdout */
            count = 0; /* Don't re-send the client's data */
			if(cli_req.command_response == SUNOEM_CLI_CMD_EOF && cli_rsp->command_response != 0 && rsp->ccode == 0) {
			cli_rsp->command_response = 1;
			}
        } while (cli_rsp->command_response == 0 && cli_rsp->buf[0] != '\0');
    }
    return ((cli_rsp->command_response == SUNOEM_SUCCESS) ? 0 : -1);
}

#define ECHO_DATA_SIZE 64
typedef struct {
    uint16_t seq_num;
    unsigned char data[ECHO_DATA_SIZE];
} __attribute__((packed)) sunoem_echo_msg_t;

/*
 * Send and receive X packets to the BMC. Each packet has a
 * payload size of (sunoem_echo_msg_t) bytes. Each packet is tagged with a
 * sequence number
 */
static int
ipmi_sunoem_echo(struct ipmi_intf * intf, int argc, char *argv[])
{
    unsigned short num, i, j;
    struct ipmi_rs *rsp;
    struct ipmi_rq req;
    sunoem_echo_msg_t echo_req;
    sunoem_echo_msg_t *echo_rsp;
    struct timeval start_time;
    struct timeval end_time;
    unsigned int total_time, resp_time, min_time, max_time;
    int received = 0;
    int transmitted = 0;
    int quiet_mode = 0;

    if (argc < 1) {
        return (1);
    }

    if (argc > 1) {
        if (*(argv[1]) == 'q') {
            quiet_mode = 1;
        }
    }

    /* Fill in data packet */
    for (i = 0; i < ECHO_DATA_SIZE; i++) {
        echo_req.data[i] = i;
    }
    /* The number of packets to send/receive */
    num = atoi(argv[0]);

    memset(&req, 0, sizeof (req));
    req.msg.netfn = IPMI_NETFN_SUNOEM;
    req.msg.cmd = IPMI_SUNOEM_ECHO;
    req.msg.data = (uint8_t *)&echo_req;
    req.msg.data_len = sizeof (sunoem_echo_msg_t);
    echo_req.seq_num = i;
    min_time = MAXINT;
    max_time = 0;
    total_time = 0;
    for (i = 0; i < num; i++) {
        echo_req.seq_num = i;
        transmitted++;
        gettimeofday(&start_time, NULL);
        rsp = intf->sendrecv(intf, &req);
        gettimeofday(&end_time, NULL);
        resp_time = ((end_time.tv_sec - start_time.tv_sec) * 1000) +
            ((end_time.tv_usec - start_time.tv_usec) / 1000);
        if ((rsp == NULL) || (rsp->ccode != 0)) {
            lprintf(LOG_ERR, "Sun OEM echo command failed. Seq # %d",
                    echo_req.seq_num);
            break;
        }
        echo_rsp = (sunoem_echo_msg_t *) rsp->data;

        /* Test if sequence # is valid */
        if (echo_rsp->seq_num != echo_req.seq_num) {
            printf("Invalid Seq # Expecting %d Received %d\n",
                   echo_req.seq_num, echo_rsp->seq_num);
            break;
        }

        /* Test if response length is valid */
        if (rsp->session.msglen  == req.msg.data_len) {
            printf("Invalid payload size for seq # %d. "
                   "Expecting %d Received %d\n",
                    echo_rsp->seq_num, req.msg.data_len, rsp->session.msglen);
            break;
        }

        /* Test if the data is valid */
        for (j = 0; j < ECHO_DATA_SIZE; j++) {
            if (echo_rsp->data[j] != j) {
                printf("Corrupt data packet. Seq # %d Offset %d\n",
                       echo_rsp->seq_num, j);
                break;
            }
        }

        /* If the for loop terminated early - data is corrupt */
        if (j != ECHO_DATA_SIZE) {
            break;
        }

        /* cumalative time */
        total_time += resp_time;

        /* min time */
        if (resp_time < min_time) {
            min_time = resp_time;
        }

        /* max time */
        if (resp_time > max_time) {
            max_time = resp_time;
        }

        received++;
        if (!quiet_mode) {
            printf("Receive %d Bytes - Seq. # %d time=%d ms\n",
                   sizeof (sunoem_echo_msg_t), echo_rsp->seq_num, resp_time);
        }
    }
    printf("%d packets transmitted, %d packets received\n", transmitted,
           received);
    if (received) {
        printf("round-trip min/avg/max = %d/%d/%d ms\n", min_time,
               total_time/received, max_time);
    }

    return (0);
}

typedef struct {
   unsigned char oem_record_ver_num;
   unsigned char major;
   unsigned char minor;
   unsigned char update;
   unsigned char micro;
   char nano[10];
   char revision[10];
   char version[40];
   /*
    * When adding new fields (using the spare bytes),
    * add it immediately after the spare field to
    * ensure backward compatability.
    *
    * e.g.   char version[40];
    *        unsigned char spare[11];
    *        int new_item;
    *    } sunoem_version_response_t;
    */
   unsigned char spare[15];
} __attribute__((packed)) sunoem_version_response_t;

/*
 * Extract the SP version data including
 * - major #
 * - minor #
 * - update #
 * - micro #
 * - nano #
 * - Revision/Build #
 */
static int
ipmi_sunoem_version(struct ipmi_intf * intf)
{
    struct ipmi_rs *rsp;
    struct ipmi_rq req;
    sunoem_version_response_t *version_rsp;

    memset(&req, 0, sizeof (req));
    req.msg.netfn = IPMI_NETFN_SUNOEM;
    req.msg.cmd = IPMI_SUNOEM_VERSION;
    req.msg.data = NULL;
    req.msg.data_len = 0;
    rsp = intf->sendrecv(intf, &req);

    if (rsp == NULL) {
        lprintf(LOG_ERR, "Sun OEM Get SP Version Failed\n");
        return (-1);
    }
    if (rsp->ccode != 0) {
        lprintf(LOG_ERR, "Sun OEM Get SP Version Failed: %d\n",
                rsp->ccode);
        return (-1);
    }

    version_rsp = (sunoem_version_response_t *) rsp->data;

    printf("Version: %s\n", version_rsp->version);

    return (0);
}

/*
 * IPMI Max string length is 16 bytes
 * define in usr/src/common/include/ami/IPMI_SDRRecord.h
 */
#define MAX_ID_STR_LEN  16
#define MAX_SUNOEM_NAC_SIZE 64
#define LUAPI_MAX_OBJ_PATH_LEN 256
typedef struct {
    unsigned char seq_num;
    char nac_name[MAX_SUNOEM_NAC_SIZE];
} __attribute__((packed)) sunoem_nacname_t;

/*
 * Retrieve the full NAC name of the IPMI target.
 *
 * The returned nac name may be larger than the payload size.
 * In which case, it make take several request/payload to retrieve
 * the entire full path name
 *
 * The initial seq_num is set to 0. If the return seq_num is incremented,
 * only the 1st 72 bytes of the nac name is returned and the caller
 * needs to get the next set of string data.
 * If the returned seq_num is identical to the input seq_num, all data
 * has been returned.
 */
static int
ipmi_sunoem_nacname(struct ipmi_intf * intf, int argc, char *argv[])
{
    struct ipmi_rs *rsp;
    struct ipmi_rq req;
    sunoem_nacname_t nacname_req;
    sunoem_nacname_t *nacname_rsp;
    char full_nac_name[LUAPI_MAX_OBJ_PATH_LEN];

    if (argc < 1) {
        return (1);
    }

    if (strlen(argv[0]) > MAX_ID_STR_LEN) {
        lprintf(LOG_ERR, "Sun OEM nacname command failed: Max size on IPMI name\n");
        return (-1);
    }
        
    nacname_req.seq_num = 0;
    strcpy(nacname_req.nac_name, argv[0]);

    full_nac_name[0] = '\0';
    while (1) {
        memset(&req, 0, sizeof (req));
        req.msg.netfn = IPMI_NETFN_SUNOEM;
        req.msg.cmd = IPMI_SUNOEM_NACNAME;
        req.msg.data = (uint8_t *)&nacname_req;
        req.msg.data_len = sizeof (sunoem_nacname_t);
        rsp = intf->sendrecv(intf, &req);

        if (rsp == NULL) {
            lprintf(LOG_ERR, "Sun OEM nacname command failed\n");
            return (-1);
        }
        if (rsp->ccode != 0) {
            lprintf(LOG_ERR, "Sun OEM nacname command failed: %d\n",
                    rsp->ccode);
            return (-1);
        }

        nacname_rsp = (sunoem_nacname_t *) rsp->data;
        strncat(full_nac_name, nacname_rsp->nac_name, MAX_SUNOEM_NAC_SIZE);

        /*
         * break out of the loop if there is no more data
         * In most cases, if not all, the NAC name fits into a
         * single payload
         */
        if (nacname_req.seq_num == nacname_rsp->seq_num) {
            break;
        }

        /* Get the next seq of string bytes */
        nacname_req.seq_num = nacname_rsp->seq_num;

        /* Check if we exceeded the size of the full nac name */
        if ((nacname_req.seq_num * MAX_SUNOEM_NAC_SIZE) >
            LUAPI_MAX_OBJ_PATH_LEN) {
            lprintf(LOG_ERR, "Sun OEM nacname command failed: invalid path length\n");
            return (-1);
        }
    }
    printf("NAC Name: %s\n", full_nac_name);
    return (0);
            
}

/* Constants used by ipmi_sunoem_getval */
#define MAX_SUNOEM_VAL_PAYLOAD 79
#define SUNOEM_REQ_VAL 1
#define SUNOEM_GET_VAL 2

#define SUNOEM_REQ_RECV 1
#define SUNOEM_REQ_FAILED 2
#define SUNOEM_DATA_READY 3
#define SUNOEM_DATA_NOT_READY 4
#define SUNOEM_DATA_NOT_FOUND 5
#define GETVAL_MAX_RETRIES 5

typedef struct {
    unsigned char cmd_code;
    char luapi_value[MAX_SUNOEM_VAL_PAYLOAD];
}  __attribute__((packed)) sunoem_getval_t;

/*
 * Return the ILOM targer property value 
 *
 */
static int
ipmi_sunoem_getval(struct ipmi_intf * intf, int argc, char *argv[])
{
    struct ipmi_rs *rsp;
    struct ipmi_rq req;
    sunoem_getval_t getval_req;
    sunoem_getval_t *getval_rsp;
    int i;

    if (argc < 1) {
        return (1);
    }

    if (strlen(argv[0]) > MAX_SUNOEM_VAL_PAYLOAD) {
        lprintf(LOG_ERR, "Sun OEM get vlaue command failed: Max size on IPMI name\n");
        return (-1);
    }
        
    strncpy(getval_req.luapi_value, argv[0], MAX_SUNOEM_VAL_PAYLOAD);

    /*
     * Setup the initial request to fetch the data.
     * Upon function return, the next cmd (SUNOEM_GET_VAL)
     * can be requested.
     */
    memset(&req, 0, sizeof (req));
    req.msg.netfn = IPMI_NETFN_SUNOEM;
    req.msg.cmd = IPMI_SUNOEM_GETVAL;
    getval_req.cmd_code = SUNOEM_REQ_VAL;
    req.msg.data = (uint8_t *)&getval_req;
    req.msg.data_len = sizeof (sunoem_getval_t);
    rsp = intf->sendrecv(intf, &req);

    if (rsp == NULL) {
        lprintf(LOG_ERR, "Sun OEM getval1 command failed\n");
        return (-1);
    }
    if (rsp->ccode != 0) {
        lprintf(LOG_ERR, "Sun OEM getval1 command failed: %d\n",
                rsp->ccode);
        return (-1);
    }

    /*
     * Fetch the data value - if it is not ready,
     * retry the request up to GETVAL_MAX_RETRIES
     */
    for (i=0; i < GETVAL_MAX_RETRIES; i++) {
        memset(&req, 0, sizeof (req));
        req.msg.netfn = IPMI_NETFN_SUNOEM;
        req.msg.cmd = IPMI_SUNOEM_GETVAL;
        getval_req.cmd_code = SUNOEM_GET_VAL;
        req.msg.data = (uint8_t *)&getval_req;
        req.msg.data_len = sizeof (sunoem_getval_t);
        rsp = intf->sendrecv(intf, &req);

        if (rsp == NULL) {
            lprintf(LOG_ERR, "Sun OEM getval2 command failed\n");
            return (-1);
        }

        if (rsp->ccode != 0) {
            lprintf(LOG_ERR, "Sun OEM getval2 command failed: %d\n",
                    rsp->ccode);
            return (-1);
        }
        
        getval_rsp = (sunoem_getval_t *) rsp->data;

        if (getval_rsp->cmd_code == SUNOEM_DATA_READY) {
            printf("Target Value: %s\n", getval_rsp->luapi_value);
            return (0);
        } else if (getval_rsp->cmd_code == SUNOEM_DATA_NOT_FOUND) {
            printf("Target: %s not found\n", getval_req.luapi_value);
            return (0);
        }

        sleep(1);
    }

    printf("Unable to retreive target value\n");

    return (-1);
}

int
ipmi_sunoem_main(struct ipmi_intf * intf, int argc, char ** argv)
{
	int rc = 0;

	if (argc == 0 || strncmp(argv[0], "help", 4) == 0) {
		ipmi_sunoem_usage();
		return 0;
	}
	if ((strcmp(argv[0], "cli") != 0)&&(strcmp(argv[0], "fan") != 0)&&(strcmp(argv[0], "led") != 0)&&(strcmp(argv[0], "sbled") != 0)&&(strcmp(argv[0], "sshkey") != 0)&&(strcmp(argv[0], "sensor") != 0)&&(strcmp(argv[0], "ping") != 0)&&(strcmp(argv[0], "version") != 0)&&(strcmp(argv[0], "nacname") != 0)&&(strcmp(argv[0], "getval") != 0))
	{
		lprintf(LOG_ERR, "Invalid sunoem command: %s", argv[0]);
		return -1;
	}

    if (strcmp(argv[0], "cli") == 0) {
        rc = ipmi_sunoem_cli(intf, argc - 1, &argv[1]);
    }

	if (strncmp(argv[0], "fan", 3) == 0) {
		if (argc < 2) {
			ipmi_sunoem_usage();
			return -1;
		}
		else if (strncmp(argv[1], "speed", 5) == 0) {
			if (argc < 3) {
				ipmi_sunoem_usage();
				return -1;
			}
			rc = ipmi_sunoem_fan_speed(intf, argv[2]);
		}
		else {
			ipmi_sunoem_usage();
			return -1;
		}
	}

	if ((strncmp(argv[0], "led", 3) == 0) || (strncmp(argv[0], "sbled", 5) == 0)) {
		if (argc < 2) {
			ipmi_sunoem_usage();
			return -1;
		}

		if (strncmp(argv[1], "get", 3) == 0) {
			if (argc < 3) {
				char * arg[] = { "all" };
				rc = ipmi_sunoem_led_get(intf, 1, arg);
			} else {
				rc = ipmi_sunoem_led_get(intf, argc-2, &(argv[2]));
			}
		}
		else if (strncmp(argv[1], "set", 3) == 0) {
			if (argc < 4) {
				ipmi_sunoem_usage();
				return -1;
			}
			rc = ipmi_sunoem_led_set(intf, argc-2, &(argv[2]));
		}
		else {
			ipmi_sunoem_usage();
			return -1;
		}
	}

	if (strncmp(argv[0], "sshkey", 6) == 0) {
		if (argc < 2) {
			ipmi_sunoem_usage();
			return -1;
		}
		else if (strncmp(argv[1], "del", 3) == 0) {
			uint8_t uid;
			if (argc < 3) {
				ipmi_sunoem_usage();
				return -1;
			}
			uid = (uint8_t)strtoul(argv[2], NULL, 0);
			rc = ipmi_sunoem_sshkey_del(intf, uid);
		}
		else if (strncmp(argv[1], "set", 3) == 0) {
			uint8_t uid;
			if (argc < 4) {
				ipmi_sunoem_usage();
				return -1;
			}
			uid = (uint8_t)strtoul(argv[2], NULL, 0);
			rc = ipmi_sunoem_sshkey_set(intf, uid, argv[3]);
		}
		else {
			ipmi_sunoem_usage();
			return -1;
		}
	}

	if (strncmp(argv[0], "sensor", 6) == 0) {
		if (argc < 2) {
			ipmi_sunoem_usage();
			return -1;
		}
		else if (strncmp(argv[1], "set", 3) == 0) {
			if (argc < 4) {
				ipmi_sunoem_usage();
				return -1;
			}
			rc = ipmi_sunoem_sensor_set(intf, argc-2, &(argv[2]));
		}
		else {
			ipmi_sunoem_usage();
			return -1;
		}
	}

	if (strcmp(argv[0], "ping") == 0) {
		if (argc < 2) {
			ipmi_sunoem_usage();
			return -1;
		}
                rc = ipmi_sunoem_echo(intf, argc-1, &(argv[1]));
        }

	if (strcmp(argv[0], "version") == 0) {
                rc = ipmi_sunoem_version(intf);
        }

	if (strcmp(argv[0], "nacname") == 0) {
		if (argc < 2) {
			ipmi_sunoem_usage();
			return -1;
		}
        rc = ipmi_sunoem_nacname(intf, argc-1, &(argv[1]));
    }

    if (strcmp(argv[0], "getval") == 0) {
		if (argc < 2) {
			ipmi_sunoem_usage();
			return -1;
		}
        rc = ipmi_sunoem_getval(intf, argc-1, &(argv[1]));
    }

	return rc;
}
